/*****************************************************************************************
 Copyright (c) 2002-2005 The UbixOS Project
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification, are
 permitted provided that the following conditions are met:

 Redistributions of source code must retain the above copyright notice, this list of
 conditions, the following disclaimer and the list of authors.  Redistributions in binary
 form must reproduce the above copyright notice, this list of conditions, the following
 disclaimer and the list of authors in the documentation and/or other materials provided
 with the distribution. Neither the name of the UbixOS Project nor the names of its
 contributors may be used to endorse or promote products derived from this software
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 $Id$

*****************************************************************************************/

#include <sys/io.h>

/************************************************************************

Function: inline unsigned char inportByte(unsigned int port);
Description: This Funciton Will Input One Byte From A Port
Notes:

************************************************************************/
unsigned char inportByte(unsigned int port) {
  unsigned char retVal;
  asm volatile(
    "inb %%dx,%%al"
    : "=a" (retVal)
    : "d" (port)
    );
  return(retVal);
  }

/************************************************************************

Function: inline unsigned char inportWord(unsigned int port);
Description: This Funciton Will Input One Word From A Port
Notes:

************************************************************************/  
unsigned short inportWord(unsigned int port) {
  unsigned short retVal;
  asm volatile(
    "inw %%dx,%%ax"
    : "=a" (retVal)
    : "d" (port)
    );
  return(retVal);
  }

/************************************************************************

Function: inline void outportByte(unsigned int port,unsigned char value);
Description: This Funciton Will Outputut One Byte To A Port
Notes:

************************************************************************/
void outportByte(unsigned int port,unsigned char value) {
  asm volatile(
    "outb %%al,%%dx"
    :
    : "d" (port), "a" (value)
    );
  }

/************************************************************************

Function: inline void outportByteP(unsigned int port,unsigned char value);
Description: This Funciton Will Outputut One Byte To A Port With A Delay
Notes:

************************************************************************/
void outportByteP(unsigned int port,unsigned char value) {
  asm volatile(
    "outb %%al,%%dx\n"
    "outb %%al,$0x80\n"
    :
    : "d" (port), "a" (value)
    );
  }  

/************************************************************************

Function: inline void outportWord(unsigned int port,unsigned char value);
Description: This Funciton Will Outputut One Word To A Port
Notes:

************************************************************************/
void outportWord(unsigned int port,unsigned short value) {
  asm volatile(
    "outw %%ax,%%dx"
    :
    : "d" (port), "a" (value)
    );
  }

/************************************************************************

Function: inline void outportDWord(unsigned int port,unsigned char value);
Description: This Funciton Will Outputut One DWord To A Port
Notes:

************************************************************************/
void outportDWord(unsigned int port,unsigned long value) {
  asm volatile(
    "outl %%eax,%%dx"
    :
    : "d" (port), "a" (value)
    );
  }

/************************************************************************

Function: inline unsigned char inportDWord(unsigned int port);
Description: This Funciton Will Input One DWord From A Port
Notes:

************************************************************************/
unsigned long inportDWord(unsigned int port) {
  unsigned long retVal;
  asm volatile(
    "inl %%dx,%%eax"
    : "=a" (retVal)
    : "d" (port)
    );
  return(retVal);
  }

/***
 END
 ***/
